// Adjacency used for framing objects
// First version: Brute force solution.

// Should use neighbor information (e.g. half-edge)
// to make the search more efficient. I will make that
// version when I find the time.

#define MAIN
#include "MicroGlut.h"
#include "GL_utilities.h"
#include "VectorUtils4.h"
#include "LittleOBJLoader.h"
#include "LoadTGA.h"
// uses framework Cocoa
// uses framework OpenGL

#define near 1.0
#define far 30.0
#define right 0.5
#define left -0.5
#define top 0.5
#define bottom -0.5
GLfloat projectionMatrix[] = {	2.0f*near/(right-left), 0.0f, (right+left)/(right-left), 0.0f,
								0.0f, 2.0f*near/(top-bottom), (top+bottom)/(top-bottom), 0.0f,
								0.0f, 0.0f, -(far + near)/(far - near), -2*far*near/(far - near),
								0.0f, 0.0f, -1.0f, 0.0f };

Model *m, *m2;
GLint *adjacencyIndex;

// Reference to shader program
GLuint program, programstd;
GLuint tex;

int findThird(Model *m, unsigned int a, unsigned int b, unsigned int avoid)
{
	for (int i = 0; i < m->numIndices; i+=3)
	{
		if (m->indexArray[i] == a || m->indexArray[i+1] == a || m->indexArray[i+2] == a)
		if (m->indexArray[i] == b || m->indexArray[i+1] == b || m->indexArray[i+2] == b)
		if (m->indexArray[i] != avoid && m->indexArray[i+1] != avoid && m->indexArray[i+2] != avoid)
		{
			for (int j = i; j < i+3; j++)
			{
				if (m->indexArray[j] != a && m->indexArray[j] != b)
				{
					// printf("Found adj %d to %d and %d (avoiding %d)\n", m->indexArray[j], a, b, avoid);
					return m->indexArray[j];
				}
			}
		}
	}
	printf("No adjacency to %d %d (avoiding %d)\n", a, b, avoid);
	return 0;
}

// Brute force adjacency builder.
// A better method would be to collect all triangles for all vertices (complete navigation information) and work from there.
// Loop through all triangles. For each vertex, count number of triangles.
// Allocate lists of triangles.
// Loop through all triangles. For each vertex, build a list of triangles.
// An alternative is to dynamically change the triangle list, with realloc or Vector push_back().
GLint *buildAdjacencyIndex(Model *m)
{
	GLint *newIndexArray = (GLint *)malloc(sizeof(GLint) * m->numIndices * 2); // 6 per triangle instead of 3

	for (int i = 0; i < m->numIndices; i+=3)
	{
		// Find any triangle with both these EXCEPT the current one!
		int c = findThird(m, m->indexArray[i], m->indexArray[i+1], m->indexArray[i+2]); // Third argument = the one to avoid
		// Put these on newIndexArray!
		newIndexArray[i*2] = m->indexArray[i];
		newIndexArray[i*2+1] = c;
		newIndexArray[i*2+2] = m->indexArray[i+1];

		c = findThird(m, m->indexArray[i+1], m->indexArray[i+2], m->indexArray[i]); // Third argument = the one to avoid
		newIndexArray[i*2+3] = c;
		newIndexArray[i*2+4] = m->indexArray[i+2];
		
		c = findThird(m, m->indexArray[i+2], m->indexArray[i], m->indexArray[i+1]); // Third argument = the one to avoid
		newIndexArray[i*2+5] = c;
	}
	return newIndexArray;
}


void init(void)
{
	// GL inits
	glClearColor(0.2,0.2,0.5,0);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_TRUE);
	printError("GL inits");

	// Load and compile shaders
	program = loadShadersG("minimal.vert", "minimalcolor.frag", "adjacency.gs");
	programstd = loadShadersG("minimal.vert", "minimallight.frag", "passthrough.gs");
	glUseProgram(program);
	printError("init shader");
	
	// Upload geometry to the GPU:
	m = LoadModel("bunny.obj"); // Model that is modified to adjacency information
	m2 = LoadModel("bunny.obj"); // Original model
//	m = LoadModel("cube.obj");
//	m2 = LoadModel("cube.obj");
	
	// Build an index array with adjecency
	adjacencyIndex = buildAdjacencyIndex(m);
	
	// Replace the index array
	glBindVertexArray(m->vao);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m->ib);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, m->numIndices*sizeof(GLuint)*2, adjacencyIndex, GL_STATIC_DRAW);

	mat4 camMatrix = lookAt(0, 0, 2, 0,0,0, 0,1,0);

	glUseProgram(program);
	glUniformMatrix4fv(glGetUniformLocation(program, "camMatrix"), 1, GL_TRUE, camMatrix.m); // A "look-at" function could be more practical
	glUniformMatrix4fv(glGetUniformLocation(program, "projMatrix"), 1, GL_TRUE, projectionMatrix);

	glUseProgram(programstd);
	glUniformMatrix4fv(glGetUniformLocation(program, "camMatrix"), 1, GL_TRUE, camMatrix.m); // A "look-at" function could be more practical
	glUniformMatrix4fv(glGetUniformLocation(program, "projMatrix"), 1, GL_TRUE, projectionMatrix);

	printError("init arrays");
}

GLfloat a = 0.0;

static void ReportAerror(const char *caller, const char *name)
{
	static unsigned int draw_error_counter = 0; 
   if(draw_error_counter < NUM_DRAWMODEL_ERROR)
   {
		    fprintf(stderr, "%s warning: '%s' not found in shader!\n", caller, name);
		    draw_error_counter++;
   }
   else if(draw_error_counter == NUM_DRAWMODEL_ERROR)
   {
		    fprintf(stderr, "%s: Number of error bigger than %i. No more vill be printed.\n", caller, NUM_DRAWMODEL_ERROR);
		    draw_error_counter++;
   }
}

void DrawModelAdj(Model *m, GLuint program, const char* vertexVariableName, const char* normalVariableName, const char* texCoordVariableName, GLint* adjIndexArray)
{
	if (m != NULL)
	{
		GLint loc;
		
		glBindVertexArray(m->vao);	// Select VAO
		glUseProgram(program); // Added 2022-03

		glBindBuffer(GL_ARRAY_BUFFER, m->vb);
		loc = glGetAttribLocation(program, vertexVariableName);
		if (loc >= 0)
		{
			glVertexAttribPointer(loc, 3, GL_FLOAT, GL_FALSE, 0, 0); 
			glEnableVertexAttribArray(loc);
		}
		else
			ReportAerror("DrawModelAdj", vertexVariableName);
		
		if (normalVariableName!=NULL)
		{
			loc = glGetAttribLocation(program, normalVariableName);
			if (loc >= 0)
			{
				glBindBuffer(GL_ARRAY_BUFFER, m->nb);
				glVertexAttribPointer(loc, 3, GL_FLOAT, GL_FALSE, 0, 0);
				glEnableVertexAttribArray(loc);
			}
			else
				ReportAerror("DrawModelAdj", normalVariableName);
		}
	
		if ((m->texCoordArray != NULL)&&(texCoordVariableName != NULL))
		{
			loc = glGetAttribLocation(program, texCoordVariableName);
			if (loc >= 0)
			{
				glBindBuffer(GL_ARRAY_BUFFER, m->tb);
				glVertexAttribPointer(loc, 2, GL_FLOAT, GL_FALSE, 0, 0);
				glEnableVertexAttribArray(loc);
			}
			else
				ReportAerror("DrawModelAdj", texCoordVariableName);
		}
		
		// Only difference from the original:
		glDrawElements(GL_TRIANGLES_ADJACENCY, m->numIndices*2, GL_UNSIGNED_INT, 0L);
	}
}


void display(void)
{
	printError("pre display");

	// clear the screen
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	a += 0.05;
	mat4 rot, trans, total;

	rot = Ry(a/2) * Rx(a/10);
	trans = T(a/100, 0, 0);
	trans = T(0, 0, 0);
	total = rot * trans;
	glUseProgram(programstd);
	glUniformMatrix4fv(glGetUniformLocation(programstd, "mdlMatrix"), 1, GL_TRUE, total.m);
	glUseProgram(program);
	glUniformMatrix4fv(glGetUniformLocation(program, "mdlMatrix"), 1, GL_TRUE, total.m);

	if (glutKeyIsDown(' '))
		DrawModel(m2, programstd, "inPosition", "inNormal", "inTexCoord");
	DrawModelAdj(m, program, "inPosition", NULL, "inTexCoord", adjacencyIndex);
		
	printError("display");
	
	glutSwapBuffers();
}

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	glutInitContextVersion(3, 2);
	glutCreateWindow ("geometry shader adjacency example");
	glutDisplayFunc(display);
	init ();
	glutRepeatingTimer(20);
	glutMainLoop();
	exit(0);
}
